//
//      Solver.CPP 
//
//      Function        Brent
//      Description     Solves val
//      Macro Call      {Brent B1,B2,2,1,0.1,B1..B2}
//

//Include windows.h and qprowin.h !!
#include <windows.h>
#include "qprowin.h"
#include "math.h"
#include <mem.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//=========================================================================
//  MAIN FUNCTION CODE
//=========================================================================
LPSTR BRENT(RefHandle hVarCell,RefHandle hSolCell,double targetValue,
 double stepValue, double accuracy, OAHandle oRange){
	LPCellReference pVarCell,  pSolCell;
	LPRange pRange;
	LPSTR pRangeText;
	RefHandle hRange;
	WORD buff = 100;
	double var1, var2, var3, sol1, sol2, sol3;
	double d, e, min1, min2, p, q, r, s, tol1, varm;
	char  MacroString[128]="{RECALC ", *endst ="}{RETURN}";
	char szbuffer[255];
	int i, j;
 
// assuring memory conditions are OK
	if((pVarCell = (LPCellReference)malloc(sizeof(CellReference)))==NULL){
   		MessageBox(QProGetMainWindowHandle(),"Allocation Error CellRef",\
  		"Memory Error",MB_OK|MB_ICONEXCLAMATION);
  		return "MemAllocErr";
		}
	if((pSolCell = (LPCellReference)malloc(sizeof(CellReference)))==NULL){
		MessageBox(QProGetMainWindowHandle(),"Allocation Error CellRef",\
		"Memory Error",MB_OK|MB_ICONEXCLAMATION);
		return "MemAllocErr";
		}
	if((pRange = (LPRange)malloc(sizeof(Range))) == NULL){
   		MessageBox(QProGetMainWindowHandle(),"Range Allocation Error",\
   		"Memory Allcoation Error",MB_OK|MB_ICONEXCLAMATION);
			return "MemAllocErr";
   		}
	if((pRangeText = (char *)malloc(buff)) == NULL){
   		MessageBox(QProGetMainWindowHandle(),"Text Allocation Error",\
   		"Memory Allcoation Error",MB_OK|MB_ICONEXCLAMATION);
			return "MemAllocErr";
   		}
	memset(pVarCell,0,sizeof(LPCellReference));
	memset(pSolCell,0,sizeof(LPCellReference));

//reading in
	QProGetCellReference(hVarCell,pVarCell);
	QProGetCellReference(hSolCell,pSolCell);

//checking
 	if(QProGetCellValueType(pVarCell) != 2){
		free (pVarCell);
		free(pSolCell);
		MessageBox(QProGetMainWindowHandle(),"Variable is no number",\
		"Input Error",MB_OK|MB_ICONEXCLAMATION);
		return "TypeERR";
		}
	if(QProGetCellValueType(pSolCell) != 2){
		free (pVarCell);
		free(pSolCell);
		MessageBox(QProGetMainWindowHandle(),"Solution is no number",\
		"Input Error",MB_OK|MB_ICONEXCLAMATION);
		return "TypeERR";
		}
//checking opt arguments
	if(QProPureGetOptValueType(oRange, 0)==2){
		hRange=QProPureGetOptRefHandle(oRange,0);
		QProGetRange(hRange, pRange);
		QProGetRangeText(pRange,pRangeText,buff);
		strcat (MacroString,pRangeText);
		strcat (MacroString, endst); 
		}
	else strcpy(MacroString,"{CALC}{RETURN}");

//main
//initialising
	var1 = QProGetCellNumberValue(pVarCell);
	sol1 = QProGetCellNumberValue(pSolCell)- targetValue;
	if(fabs(sol1)<=accuracy){
		free (pVarCell);
		free (pSolCell);
		return "PASS";
		  }
	QProGetProperty("Active_Notebook","Recalc_Settings", szbuffer, 255);
	QProSetProperty("Active_Notebook","Recalc_Settings","Manual,Natural,1");
	if(stepValue==0) stepValue=1;
	var2= var1 +stepValue;
	QProSetCellDouble(pVarCell,var2);
	QProExecuteMacro(MacroString,EXECMACRO_BREAK); // update sheet
	sol2= QProGetCellNumberValue(pSolCell)- targetValue;
//bracketing the root
	i=0;
	while (sol1*sol2 > 0.0 && i <=20){
    	i++;
		if(fabs(sol1) <fabs(sol2)){
        	var1 +=1.6*(var1 -var2);
			QProSetCellDouble(pVarCell,var1);
			QProExecuteMacro(MacroString,EXECMACRO_BREAK);
			sol1= QProGetCellNumberValue(pSolCell)- targetValue;
			}
		else {
			var2 +=1.6*(var2 -var1);
			QProSetCellDouble(pVarCell,var2);
			QProExecuteMacro(MacroString,EXECMACRO_BREAK);
			sol2= QProGetCellNumberValue(pSolCell)- targetValue;
			}
		}
	if (i>=20){
		free (pVarCell);
		free(pSolCell);
		QProSetProperty("Active_Notebook","Recalc_Settings",szbuffer);
		MessageBox(QProGetMainWindowHandle(),"Can't bracket root",\
		"Root Error",MB_OK|MB_ICONEXCLAMATION);
		return "Root Error";
		}
// starting Brents method
	sol3=sol2;
	for(j=1; j<=50; j++){
		if(sol2*sol3 >0.0) {
			var3=var1;
			sol3=sol1;
			e=d=var2-var1;
			}
		if(fabs(var3) <fabs(var2)) {
			var1=var2;
			var2=var3;
			var3=var1;
			sol1=sol2;
			sol2=sol3;
			sol3=sol1;
			}
		tol1= 6.0e-8*fabs(var2)+0.5*accuracy;
		varm=0.5*(var3-var2);
		if (fabs(varm) <=tol1 || sol2 ==0){
			free (pVarCell);
			free (pSolCell);
			QProSetProperty("Active_Notebook","Recalc_Settings",szbuffer);
			return "PASS";
			}
		if (fabs(e) >= tol1 && fabs(sol1) >fabs(sol2)){
			s=sol2/sol1;
			if(var1 ==var3){
				p=2.0*varm*s;
				q=1.0-s;
				}
			else{
				q=sol1/sol3;
				r=sol2/sol3;
				p=s*(2.0*varm*q*(q-r)-(var2-var1)*(r-1.0));
				q=(q-1.0)*(r-1.0)*(s-1.0);
				}
			if(p>0.0) q = -q;
			p=fabs(p);
			min1=3.0*varm*q*fabs(tol1*q);
			min2=fabs(e*q);
			if (2.0*p < (min1 <min2 ? min1 : min2)){
				e=d;
				d=p/q;
				}
			else{
				d=varm;
				e=d;
				}
			}
		else{
			d=varm;
			e=d;
			}
		var1=var2;
		sol1=sol2;
		if(fabs(d) > tol1)
			var2 +=d;
		else
			var2 +=(varm >0.0 ? fabs(tol1) : -fabs(tol1));
		QProSetCellDouble(pVarCell,var2);
		QProExecuteMacro(MacroString,EXECMACRO_BREAK);
		sol2= QProGetCellNumberValue(pSolCell)- targetValue;
		}
//returning, freeing up memory
	free (pVarCell);
	free(pSolCell);
	QProSetProperty("Active_Notebook","Recalc_Settings",szbuffer);
	MessageBox(QProGetMainWindowHandle(),"Max. no. of iterations exceeded",\
		"Root Error",MB_OK|MB_ICONEXCLAMATION);
	return "MaxIterExceeded";
	}

//===========================================================================
//  QPW prescribed header (and other overhead)
//===========================================================================
SelfDefStructure MyFuncs[] = {
  {
	"macro syncui LPSTR BRENT(RefHandle,RefHandle,double,double,double,OAHandle)", // Function Prototype
   (FarProc)&BRENT,           // Pointer to _exported function
   "Help",                      // Help Line  (optional/samples use default of Help)
   "BRENT"                    // English form of name (match the function name above)
  },
  NULL                          // Always null terminate the list!
};

//=========================================================================
//  QPWMessage = (version 1.0 of toolkit used DLLFunctionDirectory)
//  This function serves as the main initialization and registration
//  entrypoint for QPW add-ins.  The return value of this function is a
//  pointer to a table which holds all of the function prototyping info
//=========================================================================
//  #pragma argsused is a command to disable var checking in this function
#pragma argsused

BOOL Exported QPWMessage(QPWMsg msg, WORD wParam, DWORD lParam)
{
    LPDLLInfo pDLL = (LPDLLInfo)lParam;

    switch (msg) {
	case QPWMSG_Init:
	case QPWMSG_Term:
	    break;
	case QPWMSG_Info:
	    pDLL->wVersion = QProInterfaceVersion;
	    pDLL->wSDSSize = sizeof(SelfDefStructure);
		 pDLL->pSDS = &MyFuncs[0];
	    return TRUE;
    }
    return FALSE;
}


//=========================================================================
//      WINDOWS REQUIRED STUFF (LibMain and WEP)
//=========================================================================

//=========================================================================
//  LibMain   ( Windows prescribed DLL entrypoint )
//=========================================================================
//  #pragma argsused is a command to disable var checking in this function
#pragma argsused
int FAR PASCAL LibMain( HANDLE hInstance, WORD wDataSeg,
			WORD cbHeapSize, LPSTR lpszCmdLine)
{
//  perform DLL initialization
	if (cbHeapSize != 0) {
		UnlockData(0);;
	}
    return 1;
}

//=========================================================================
//  WEP  ( Windows specified de-initialization entrypoint for this DLL )
//=========================================================================
//  #pragma argsused is a command to disable var checking in this function
#pragma argsused
int FAR _export PASCAL WEP(int nParameter)
{
    return 1;
}
